/*******************************************************************/
/*                                                                 */
/*   File name :  usbadc2k5.h                                      */
/*                                                                 */
/*      Version:   2015-10-20                                      */
/*                                                                 */
/*   Copyright (c) 2015-     T.Kondo/NICT All Right Reserved       */
/*                                                                 */
/*******************************************************************/

#include <stdio.h>
#include <signal.h> 
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifndef WIN32
#include <asm/byteorder.h>   
#endif
#ifdef WIN32
  #define __LITTLE_ENDIAN_BITFIELD       //  Intel CPU
#endif
#include "lib_from_K5.h"

#ifdef __CYGWIN__
#ifndef linux
#define linux
#endif
#endif

#ifdef linux
#include <sys/time.h>
#endif

#ifndef WIN32
#include <unistd.h>
#endif



#ifdef linux
 #ifdef _K5_USE_FOPEN64_
  #ifndef _FOPEN_DEFINED_
   #define _FOPEN_DEFINED_
   #define fopen(_Z,_X) fopen64(_Z,_X)
  #endif
 #endif
 #ifdef _LARGEFILE_SOURCE
  #ifndef _FSEEK_DEFINED_
   #define _FSEEK_DEFINED_
   #define fseek(_Z,_X,_W) fseeko(_Z,_X,_W)
  #endif
  #ifndef _FTELL_DEFINED_
   #define _FTELL_DEFINED_
   #define ftell(_Z) ftello(_Z)
  #endif
 #endif
#endif


/* for WIN32 */

#ifdef WIN32
#pragma warning (disable:4996)
#ifndef _FORWIN32SET_
 #define _FORWIN32SET_
 typedef unsigned int    u_int32_t;
 typedef unsigned int    uint32_t;
 typedef unsigned short int    u_int16_t;
 typedef unsigned short int    uint16_t;
 typedef unsigned char   u_int8_t;
 typedef _int64 int64_t;
 #define popen(_Z,_X) _popen(_Z,_X)
 #define pclose(_Z) _pclose(_Z)
 #define fseek(_Z,_X,_W) _fseeki64(_Z,_X,_W)
 #define ftell(_Z) _ftelli64(_Z)
 #endif
#endif


#ifndef USE_ASM
#define USE_ASM
#endif


/* constant set */
#ifndef _SET_K5_CONSTANTS_
#define _SET_K5_CONSTANTS_

#define HEAD_SEARCH_RANGE 600  // header serach range (in 32bit word) -0.5*HEAD_SEARCH_RANGE to +0.5*HEAD_SEARCH_RANGE
#define VDIF_KASHIMA_SYNC 0xACABFEEDUL    // sync pattern for VDIF Kashima format

#ifndef FILEMAXSIZE
#define FILEMAXSIZE  2000000000   //  File maximum limit size in bytes when limiting
#endif

#ifndef PIRA
#define PIRA   0.017453292519943295769236907684886    /* PI/180  degree to radian */
#endif
#ifndef PIDE
#define PIDE   57.295779513082320876798154814105      /* 180/PI radian to degree  */
#endif
#ifndef TWOPI
#define TWOPI  6.283185307179586476925286766559       /* 2PI                      */
#endif
#ifndef PI
#define PI     3.1415926535897932384626433832795      /* PI                       */
#endif
#ifndef HALFPI
#define HALFPI 1.5707963267948966192313216916398      /*  PI/2                    */
#endif
#ifndef VLIGHT
#define VLIGHT 2.99792458E+8    /* speed of light (m/sec)  */
#endif

/* mainly used in libapri etc */
#ifndef CONVA
#define CONVA 4.8481368110953599358991410235795e-6    /* PI/(3600*180)  arcsec to radian */
#endif
#ifndef CONVS
#define CONVS 7.2722052166430399038487115353692e-5    /* PI/(3600*12) time sec to radian */
#endif
#ifndef CONVD
#define CONVD 0.017453292519943295769236907684886    //*PI/180  degree to radian  */
#endif
#ifndef CONVH
#define CONVH 0.26179938779914943653855361527329    /* PI/12 hour to radian  */
#endif


#define REWIND_BYTE 200       // for K5/VSSP access

#endif   // *** END of _SET_K5_CONSTANTS_


#define int64_t long long int     // to avoid format "%lld" warning in linux


typedef struct runp     // run parameters for this
{
    int ptype;   // run parameter type  =1  mode1,  =2 mode 2
    int channel;
    int group;
    int soffset;
    int period;
    int reverse;
    int usbadc_headlen;  // USBADC headder length (bytes)
    int usbadc_nsamples; // USBADC # of samples in a frame
	int64_t skip_bytes;   // skip bytes (1st skip over # of bytes)
    int scan;    // scan# for get MODE from VEX file (default=1) 
    int odd;    // =1 4byte integer offset mode in case of 64 track
    int monit;   // =1  inofo_file monitor ON,  =0 no monitoring
    int tflag;
    int itimes[5]; // data start time
    char usbadc_fname[256];
    char k5name[256];
    char k5name_given[256];
    char outdir[256];
    char info_file[256];
    char vex_file[256];
    char sid[4];
    char dflt_pass[4];
    // for mode2 only run parameres are as follows
    int numch;
    int ntrack;
    int adbit;    // AD bits in K5 format
    int fanout;
    int parity;
    int nrzm;
    int vlba;
    int ibit[32];
    int tot_track;
	int k5numch; 
	int e2bit;
	int m5ch;
	int chtable[4];
	int nbitstream;
    double fmhz;
    double fact;
    int flag_compulsory;
} runp_t;



typedef struct timeval_vdif    // time related variables used in vdif utilities
{
	int sec_fm_ref;   // Seconds from reference epoch
	int ref_epo;      // Reference epoch for second count.  

    int year;      // 4 digit year
    int ddd;       // day of year
    int hh;        // hour
    int mm;        // minute
    int ss;        // sec in minute
    int second;    // sec in day
    double dtime;  // time in day (ddd+second/86400.0)
    int month;     // month
    int mday;      // day of month
	int mjd;       // MJD
	double fday;   // fractional part of day
} timeval_vdif_t;


/* VSSP32 header structure for AUX FORMAT #22 */
struct vssp32h {
#ifdef WIN32
  #pragma pack(4)
#endif
	u_int32_t syncptn1;		/* 0xffffffff */
#if defined(__BIG_ENDIAN_BITFIELD)
	u_int32_t syncptn2 : 8;
	u_int32_t bitwidth : 2;
	u_int32_t fsample : 4;
	u_int32_t chmode : 1;
	u_int32_t seconds : 17;

	u_int32_t major : 4;
	u_int32_t minor : 4;
	u_int32_t afsize : 8;
	u_int32_t eflag : 1;
	u_int32_t year : 6;
	u_int32_t totalday : 9;

	u_int32_t fsample2 : 16;
	u_int32_t lpf  : 8;
	u_int32_t affmt: 8;
#ifdef WIN32
  #pragma pack(2)
#endif
	u_int16_t adbit: 8;
	u_int16_t numch: 8;
    char aux[14];
#endif
#if defined(__LITTLE_ENDIAN_BITFIELD)
	u_int32_t seconds : 17;
	u_int32_t chmode : 1;
	u_int32_t fsample : 4;
	u_int32_t bitwidth : 2;
	u_int32_t syncptn2 : 8;

	u_int32_t totalday : 9;
	u_int32_t year : 6;
	u_int32_t eflag : 1;
	u_int32_t afsize : 8;
	u_int32_t minor : 4;
	u_int32_t major : 4;

	u_int32_t affmt: 8;
	u_int32_t lpf  : 8;
	u_int32_t fsample2 : 16;

#ifdef WIN32
  #pragma pack(2)
#endif
	u_int16_t numch: 8;
	u_int16_t adbit: 8;

    char aux[14];
#endif

	/* Continue here version number and aux fields */
#ifdef WIN32
};
#else
} __attribute__((__packed__));
#endif
typedef struct vssp32h vssp32h_t;
#ifdef WIN32
  #pragma pack()
#endif


typedef struct vssp32     // vssp32 data structure
{
	double fsample;   // sampling frequency (Hz)
	int adbit;      // AD resolution
	int numch;  // # of channels
    int af_format;  // AUX field format #
    int year;       // yaear
    int ddd;        // total day
    int seconds;    // second from 0h UTC
    int nsample;    // # of samples corresponding to USBADC
    int nword;
} vssp32_t;


typedef struct vdif_header     // vdif file header structure
{
	int sec_fm_ref;   // Seconds from reference epoch
	int legacy;       // Legacy mode  0: standard 32 byte VDIF data frame header
	                  //              1: legacy header-length' mode ; words 4-7 omitted from header
	int invalid;      // Invalid data   0: valid  1: invalid
	int frame;        // Data frame # within second, starting at zero; must be integral number od data drames per second
	int ref_epo;      // Reference epoch for second count.  This contains the 6-month period in which the VDIF
	                  // clock was set, with an origin of at 00UTC 1 Jan 2000, such that '0' corresponds to the
	                  // first 6 months of 2000, '1' corresponds to the 6 months starting 00h 1 July 2000, etc. 
	                  // This field rolls over to 0 again at the 6-month period starting 00h 1 Jan 2032.
	int frame_len64;    // Data fram length (including header) in units of 8 bytes (must be multiple of 8 bytes, 
	                  // max is 2^27 bytes)
	int log2_chs;     // log_2(#channels in data array). #chans must be power of 2
	int ver;           // VDIF version number
	unsigned short station_id;   //  station ID (either A2 of unsigned 16-bit integer)
	char sid[3];      // station ID (A2)
	int thread_id;    // Thread ID (0-1023)
	int bpsampl;      // bits per sample (max 32). If complex type, this is set according to the #bits in each complex-
	                  // component
	int d_type;       // data type  0: real data   1: complex data
	int edv;          // Extended data version
	int w1;           // extented data word#1 (3 bytes are valid)
	int w2;           // extended data word#2
	int w3;           // extended data word#3
	int w4;           // extended data word#4
	// below is extended format of NICT's
	int uflag;        // unit of sampling rate 0: kHz,  1: MHz
	int srate;        // sampling rate
	int sync_block;   // sync block  = w2
	char das[9];      // DAS/station name1
    timeval_vdif_t t; // time information converted to YYYY Month Day hour minute second

} vdif_header_t;

typedef struct ch_data     // ch_data variable definition
{
    short int *idata;
    int idata_over[32];
    int usampl;
} ch_data_t;

typedef struct usbadc_file     // vdif_file type variable definition
{
    /*
        Ver. 2014-04-25
         and previous one.
    */
	char fname[128]; //data file name opened or to be opened  **K5**
	FILE *fin;
	int64_t skip_bytes;   // skip bytes (1st skip over # of bytes)
	int64_t pos;   // current file pointer
	int64_t bytes1sec;  // # of bytes in 1 second including headder portion
	                    // (usually used for 1 st forward)
	int64_t f_size;  // opened file size in bytes
	int64_t pos_1st_header;  // position of 1 st correct header (valid when flag_vdif_anal=1)
	int64_t tot_frame; // total frame counter
	//int64_t tbytesread;   // total # of bytes read
    vdif_header_t vhead;
	int flag_vdif_anal;  //  flag to show vdif file format is analyzed
	double tnframe;   // total number of frames (double) (computed from file size)
    double fbps;      // supposed data rate (bps) per thread
	double fsample;   // sampling frequency (Hz)
	double tsample;   // sampling period (sec)
    double pframe;    // frame period (sec)
	double dlength;   // data length in seconds
	int headlen;      // header length in 32bit word
	int nframsec;     // frames/sec per thread
	int adbit;      // AD resolution
	int numch;  // # of channels
	int nthread; // # of threads
	int frame_bytes; // # of bytes in a frame (excepting header)
	//double frame_bytes; // # of bytes in a frame (excepting header)
	double unitp;   // usampl*tsampl unit length in sec
    u_int32_t head[8];      // for reading header portion
    u_int32_t *ubuf;        // for data field reading
    int nword;       // ubuf size (in 32bit word) allocated or to be allocated
	int nsample;     // # of samples in ubuf(frame)
	int imax;        // # of ubuf read for 1 frame
	int imax_c;      // counter to compare with imax (already read number)
	int upos;  // next read (decode) position in ubuf array
	int wordread;     // actually read words (ubuf)
    //int *idata;        // pointer to decoded data (integer) container by channel
    //double *ddata;      // pointer to decoded data (R*8) container
	//int idata_over[32]; // to contain overflowed decode data
    ch_data_t *ch;   // decode data by channel (max numch)
	int smode;     // search mode
	int usampl;      // usampl*numch is the size of idata and data
	int usampl_c;    // counter used for actual number of samples decoded data 
	int n_usampl;    // # of unit samples (starting from 1) 
	double dtime;  // time at begin of unit
	double dtimest;  // time at begin of file data  (in day)
	int npos;    //  valid number in idata or data
	int flag_idata_over;  // flag to show how many samples are over from idata  0: ok  +ve: overed # 
	int flag_read_data;      // data read flag
	int flag_read_header;   // header read flag
	int flag_unit_read_ok;  // 1 unit data read OK flag
	int flag_compulsory;  // 1: set header format to Kashima compulsory (debug purpose)
	                      // +VE (>1) : set bytes/frame compulsory with Kashima Format 
	int flag_open;        // file open flag  0: closed  1:opened
	//int flag_elec;       // set ELEC VDIF format compulsory (debug purpose)


} usbadc_file_t;


