/************************************************************************/
/*                                                                      */
/*   Program Name :  vdif2m5b                                           */
/*                                                                      */
/*     Vdif2m5b is VDIF to Mark5B file format converter                 */
/*                                                                      */
/*     Enviroment variables                                             */
/*           M5DIR      ---   default directory for Mark5 data          */
/*           M5VEX      ---   default directory for Mark5 vex file      */
/*                                                                      */
/*            Ver 1.00  2016-12-22 start development based on vdif2k5   */
/*                                                                      */
/*      Copyright (c) 2016-     T.Kondo All Right Reserved              */
/*                                                                      */
/************************************************************************/
/*

    only 1st thread ID is tracked

     2016-12-20  -- development started
     2016-12-22  -- Ver.1.00

   how to run
          vdif2m5b vdif_filename [options]
    
     where   vdif_filename -- VDIF data file name
              options (any order)

               -fsampl fmhz   -- channel sampling frequency (MHz) to compulsory set
			   -2bit mode     -- set 2bit AD decode table
			                         1: offset binary   0,1,2,3 (no conversion)
									 2: Mark V          0,2,1,3  (default for Mark5B)
									 3: signed integer  1,0,2,3

               -o m5b_name    -- Mark5B name to be created
                                 (default : see below)
               -d m5b_dir     -- directory for Mark5B file out
                                 (default  : vdif file directory)
               -s[offset] soffset
			         +VE : start offset (sec) (default is 0)
					 -VE : absolute value means skip over bytes 
               -k[ashima] [bytesframe] 
                     set header format to Kashima compulsory 
                     bytesframe: +VE bytes/frame set compulsory
               -type0   --- naming type 0  vdifname + '.m5b'
               -type1   --- naming type 1  (exchange file extention to 'm5b'

               -ch1 nn       -- set VDIF channel corresponding to Mark5 CH #1
                                     nn is VDIF channel number (1-16)
               -ch2 nn       -- set VDIF channel corresponding to Mark5 CH #2
                   ...........................
               -ch16 nn      -- set VDIF channel corresponding to Mark5 CH #16

               -chall n1:n2:....:n16    -- set VDIF channels corresponding to Mark5B all channels
                               n1  -- VDIF channel number (1-16) corresponding to Mark5B CH 1
                               n2  -- VDIF channel number (1-16) corresponding to Mark5B CH 2
                                 ...............
                               n16 -- VDIF channel number (1-16) corresponding to Mark5B CH 16

               -pid thread_id  -- pickup thread id (default is thread id encountered 1st)
  
  Naming rule for Mark5B file created (VDIF.vdif : original VDIF name)
            Type 0:  VDIF.vdif.m5b 
            Type 1:  VDIF.m5b
*/

#define PROGRAM_VERSION "vdif2m5b (Ver 1.00  2016-12-22)"

#include "vdif2m5b.h"

#define M5DIR "M5DIR" // the environment variable for Mark5 data default directory


#define DFLT_DATA_FILE "vdif.dat"
#define BUFMAX 25000   // maximum read buffer size
#define DFLT_CH_TABLE {1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16}  // default CH allocation table

int cmnd_get_num(int argc, char *argv[], char *command, char *param, int *num)
// get argument like -command parameter
// search "-command" in argument, if it is exsted, 0 is returned as
// a function value, and next argument is returned as a parameter (ascii).
// if "-command" is not exsited, -1 is returned as a function value.
// if "-command" is not accompanied by real parameter, NULL character
// is returned as param.
// 
{
    int n;
    char cbuf[80];

    param[0]='\0';

    //search command
    //printf("%s\n",command); // debug
    n=2;
    while(n <=argc) {
        strcpy(cbuf,*(argv+n-1));
        if(strncmp(cbuf,command,strlen(command))==0){  
            n++;
			*num=n;
            if(n > argc) return 0;
            strcpy(param,*(argv+n-1));
			*num++;
            return 0;
        }
        n++;
    }
    return -1;
}

int vex_parse(char *cbuf, vparam_t *vparam)  //2003-03-21
{
/*
    Parse VEX parameters by '=',':',';', and ' '.
    '=',':',';' are returned as a parsed item for further analysis
    Total number of items including '=',':',';' is defined in
    type define statesment of vparam as

    typedef struct vparam
    {
        int  n;          //number of param
        char param[60][128];   // 60x128 characters
    } vparam_t;

    In this case 60 is a maximum number of items and
    128 characters are maximum char in an item.

    If runtime error occurrs, please try to increase the number of
    items in the vparam definition.


*/

    // vex_parse =0;  single statement in a cbuf
    //           =1;  more statement left in a cbuf, start point in cbuf 
    //                is updated 
    char abuf[500];
    char *cset=" ";
    char *p=NULL;
    int i,ipos,k,kpos,flag,mflag,len;

    len=strlen(cbuf);
    //printf("len = %d\n",len);

    //remove extra space in front

    if(cbuf[0]==' '){
        kpos=0;
        for(k=0; k<len; k++){
            if(cbuf[k]!=' '){
                kpos=k;
                break;
            }
        }
        i=0;
        for(k=kpos; k<len; k++){
            cbuf[i]=cbuf[k];
            i++;
        }
        cbuf[i]='\0';
    }
    // len again

    len=strlen(cbuf);


    // parse start

    if(cbuf[0] == '*') { // comment
        vparam->n=0; 
        return 0;
    }
    i=0;
    ipos=0;
    flag=0;
    for(k=0; k<len; k++){
        if(cbuf[k] == ' ' || cbuf[k] == '=' || cbuf[k] == ':'){
            if(flag==1) {  // one parameter end
                vparam->param[ipos][i]='\0';
                flag=0;
                i=0;
                ipos++;
            }
            if(cbuf[k] != ' '){
                vparam->param[ipos][i]=cbuf[k];
                vparam->param[ipos][i+1]='\0';
                flag=0;
                i=0;
                ipos++;
            }
            goto nextk;
        }
        if(cbuf[k] == ';'){
            kpos=k;
            goto exithere;
        }
        if(cbuf[k] != ' '){
            flag=1;
            vparam->param[ipos][i]=cbuf[k];
            i++;
        }
nextk:;
    }
    kpos=k;
exithere:
    vparam->param[ipos][i]='\0';
    ipos++;
    vparam->n=ipos;
    i=0;
    mflag=0;
    if(kpos < len-1){
        for(k=kpos+1; k<len; k++) {
            if(cbuf[k] != '\0'){
                abuf[i]=cbuf[k];
                i++;
            }
            if(cbuf[k] == ';') mflag=1;
        }
    }
    if(mflag==1){
        abuf[i]='\0';
        strcpy(cbuf,abuf);
        return 1;
    }
    return 0;
}


void param_get_vdif2m5b(int argc, char *argv[],runp_t *runp)                       
{
    int n,i;
    char cbuf[256];
    int ch_table_dflt[]=DFLT_CH_TABLE;
    vparam_t vp;

	// initialize parameters

    /* common parameters */
    runp->ptype=1;
    runp->vdifname[0]='\0';
    //runp->k5name[0]='\0';
    runp->m5name[0]='\0';
    runp->m5name_given[0]='\0';
    runp->soffset=0;
    runp->period=0;
    runp->odd=0;

    /* mode 1 parameters */
    runp->channel=0;
    runp->group=0;
    runp->unit=0;
    runp->reverse=0;
    runp->info_file[0]='\0';
    runp->outdir[0]='\0';
    runp->vex_file[0]='\0';
    runp->sid[0]='\0';
    runp->scan=1;
    runp->monit=0;
	runp->flag_compulsory=0;
	runp->skip_bytes=0;
	runp->k5numch=1; // 
    runp->vdifch=1;
    runp->naming_type=0;
    runp->flag_ch_alloc=0;
    runp->pickup_tid=-1;

    /* mode 2 parameters */
    runp->numch=0;
    runp->adbit =0;
    //runp->fmhz  =(double)M5SAMPLE;
    runp->fmhz  =0.0;
	runp->nbitstream=0;
    for(n=0; n<32; n++) runp->ibit[n]=0;
    runp->fmode=3;   // default conversion file mode 1: 1ch  2: 4ch  3: 4x4ch  4: 2x4ch
    //for(n=0; n<4; n++) runp->chtable[n]=0;
    for(n=0; n<16; n++) runp->chtable[n]=ch_table_dflt[n];
    runp->e2bit=2;     // for VDIF -> Mark5B

    //runp->k5nfile=0; //  0 means auto mode  2016-08-08

    /* 1st parameter (vdif file name) get */
    if (argc>=2){
       strcpy(runp->vdifname,*(argv+1));
	   if (strlen(*(argv+1))==1 && runp->vdifname[0]=='0'){  // set default name
          strcpy(runp->vdifname,DFLT_DATA_FILE);
	   }
    }
     /* check VDIFfile name is acompanied by directory or not */
    get_true_path(runp->vdifname, M5DIR, "./",cbuf);
    strcpy(runp->vdifname,cbuf);

    /* command parameters get */
    if(cmnd_get_p(argc,argv,"-o",cbuf)==0) strcpy(runp->m5name_given,cbuf);  // Mark5 name
	if(cmnd_get_p(argc,argv,"-d",cbuf)==0){              // out dir
		strcpy(runp->outdir,cbuf);  
        /* env variable is substituted by real one */
        if((i=substitute_env_path_name(runp->outdir, cbuf))>0){
           strcpy(runp->outdir,cbuf);
        }
	}
//    if(cmnd_get(argc,argv,"-c",cbuf)==0) runp->channel = atoi(cbuf);  // CH
//    if(cmnd_get(argc,argv,"-g",cbuf)==0) runp->group   = atoi(cbuf);  // group
    //if(cmnd_get_p(argc,argv,"-s",cbuf)==0) runp->soffset = atoi(cbuf);  // start offset
    if(cmnd_get_p(argc,argv,"-p",cbuf)==0) runp->period  = atoi(cbuf);  // period
	if(cmnd_get_p(argc,argv,"-r",cbuf)==0) runp->reverse=1;  // bit reverse mode
	if(cmnd_get_p(argc,argv,"-si",cbuf)==0) strcpy(runp->sid,cbuf);  // station id
	if(cmnd_get_p(argc,argv,"-scan",cbuf)==0) {  // scan
		  runp->scan=atoi(cbuf);
          if(runp->scan < 1) runp->scan=1;
	}
	if(cmnd_get_p(argc,argv,"-monit",cbuf)==0) runp->monit=1;  // monitor mode on
    if(cmnd_get_p(argc,argv,"-fs",cbuf)==0) runp->fmhz  = atof(cbuf);  // VDIF sampling frequency
    if(cmnd_get_p(argc,argv,"-ad",cbuf)==0) runp->adbit = atoi(cbuf);  // VDIF AD bits
    if(cmnd_get_p(argc,argv,"-bs",cbuf)==0) runp->nbitstream = atoi(cbuf);  // bit stream

    if(cmnd_get(argc,argv,"-2bit",cbuf)==0) runp->e2bit = atoi(cbuf);   // 2bit encode mode
	if(cmnd_get_p(argc,argv,"-s",cbuf)==0){
		if(atof(cbuf) >= 0.0){
		    runp->soffset = atoi(cbuf);   // second offset
		} else {
			runp->skip_bytes=-atoi(cbuf);  // skip bytes
		}
	}
    if(cmnd_get_p(argc,argv,"-k",cbuf)==0){ // set compulsory mode
		runp->flag_compulsory=1;
        if(cbuf[0] != '\0'){
            if(cbuf[0] != '-') runp->flag_compulsory=atoi(cbuf);
        }
    }
    if(cmnd_get_p(argc,argv,"-type1",cbuf)==0) runp->naming_type=1;
    if(cmnd_get_p(argc,argv,"-pid",cbuf)==0) runp->pickup_tid = atoi(cbuf);   // pickup thread ID

    if(cmnd_get_p(argc,argv,"-ch1",cbuf)==0) runp->chtable[0]=atoi(cbuf);  // CH1 set
    if(cmnd_get_n(argc,argv,"-ch01",cbuf)==0) runp->chtable[0]=atoi(cbuf);  // CH1 set
    if(cmnd_get_n(argc,argv,"-ch2",cbuf)==0) runp->chtable[1]=atoi(cbuf);  // CH2 set
    if(cmnd_get_n(argc,argv,"-ch02",cbuf)==0) runp->chtable[1]=atoi(cbuf);  // CH2 set
    if(cmnd_get_n(argc,argv,"-ch3",cbuf)==0) runp->chtable[2]=atoi(cbuf);  // CH3 set
    if(cmnd_get_n(argc,argv,"-ch03",cbuf)==0) runp->chtable[2]=atoi(cbuf);  // CH3 set
    if(cmnd_get_n(argc,argv,"-ch4",cbuf)==0) runp->chtable[3]=atoi(cbuf);  // CH4 set
    if(cmnd_get_n(argc,argv,"-ch04",cbuf)==0) runp->chtable[3]=atoi(cbuf);  // CH4 set
    if(cmnd_get_n(argc,argv,"-ch5",cbuf)==0) runp->chtable[4]=atoi(cbuf);  // CH5 set
    if(cmnd_get_n(argc,argv,"-ch05",cbuf)==0) runp->chtable[4]=atoi(cbuf);  // CH5 set
    if(cmnd_get_n(argc,argv,"-ch6",cbuf)==0) runp->chtable[5]=atoi(cbuf);  // CH6 set
    if(cmnd_get_n(argc,argv,"-ch06",cbuf)==0) runp->chtable[5]=atoi(cbuf);  // CH6 set
    if(cmnd_get_n(argc,argv,"-ch7",cbuf)==0) runp->chtable[6]=atoi(cbuf);  // CH7 set
    if(cmnd_get_n(argc,argv,"-ch07",cbuf)==0) runp->chtable[6]=atoi(cbuf);  // CH7 set
    if(cmnd_get_n(argc,argv,"-ch8",cbuf)==0) runp->chtable[7]=atoi(cbuf);  // CH8 set
    if(cmnd_get_n(argc,argv,"-ch08",cbuf)==0) runp->chtable[7]=atoi(cbuf);  // CH8 set
    if(cmnd_get_n(argc,argv,"-ch9",cbuf)==0) runp->chtable[8]=atoi(cbuf);  // CH9 set
    if(cmnd_get_n(argc,argv,"-ch09",cbuf)==0) runp->chtable[8]=atoi(cbuf);  // CH9 set
    if(cmnd_get_n(argc,argv,"-ch10",cbuf)==0) runp->chtable[9]=atoi(cbuf);  // CH10 set
    if(cmnd_get_n(argc,argv,"-ch11",cbuf)==0) runp->chtable[10]=atoi(cbuf);  // CH11 set
    if(cmnd_get_n(argc,argv,"-ch12",cbuf)==0) runp->chtable[11]=atoi(cbuf);  // CH12 set
    if(cmnd_get_n(argc,argv,"-ch13",cbuf)==0) runp->chtable[12]=atoi(cbuf);  // CH13 set
    if(cmnd_get_n(argc,argv,"-ch14",cbuf)==0) runp->chtable[13]=atoi(cbuf);  // CH14 set
    if(cmnd_get_n(argc,argv,"-ch15",cbuf)==0) runp->chtable[14]=atoi(cbuf);  // CH15 set
    if(cmnd_get_n(argc,argv,"-ch16",cbuf)==0) runp->chtable[15]=atoi(cbuf);  // CH16 set

    if(cmnd_get_n(argc,argv,"-chall",cbuf)==0) { // all channel set
        vex_parse(cbuf, &vp);
        i=0;
        for(n=0; n<vp.n; n++) {
           //printf(" n= %d  %s\n",n,vp.param[n]);// debug
            if(vp.param[n][0]!=':'){
                runp->chtable[i]=atoi(vp.param[n]);
                i++;
            }
        }

    }
    // check necessity of channel re-allocation
    for(n=0; n<16; n++){
        if(runp->chtable[n] != n+1) runp->flag_ch_alloc=1;
    }
}

void print_usage(void)
{
    printf("%s\n",PROGRAM_VERSION);
    printf("\n");
    printf("    vdif2m5b vdif_file [options]\n");
    printf("\n");
    printf("where   vdif_file  -- VDIF data file name\n");
    printf("      options (any order)\n");
    printf("        -o m5bname     -- Mark5B file name to be created\n");
    printf("                          (default : see below)\n");
    printf("        -d m5dir      -- directory for Mark5B file out\n");
    printf("                          (default  : same as VDIF file directory)\n");
    //printf("        -bs nbitsream -- set Mark5B # of bit-stream (1,2,4,8,16,32)\n");
    printf("        -fsampl fmhz  -- channel sampling frequency (MHz) to compulsory set\n");
//    printf("        -u unit       -- pickup unit# (1-4) for 4 ch conversion mode\n");
//    printf("                         (-1ch option is stronger than this option)\n");
//    printf("                         (default 4x4ch mode)\n");
//    printf("        -1ch vdifch   -- set K5 data to 1ch mode and \n");
//    printf("       	                    pick up VDIF ch (starting at 1)\n");
//    printf("        -4ch ch1 ch2 ch3 ch4\n");
//    printf("                      -- set K5 data to 4 ch mode and  \n");
//    printf(" 	                       pick up VDIF 4 channels\n");
    printf("        -2bit mode    -- set 2bit AD decode table\n");
    printf("                           1: offset binary   0,1,2,3 (no conversion)\n");
    printf("                           2: Mark V standard 0,2,1,3 (default)\n");
    printf("                           3: signed integer  1,0,2,3\n");
    printf("        -s[offset] soffset\n");
    printf("                        +VE : start offset (sec) (default is 0.0)\n");
//    printf("                        -VE : absolute value means skip over bytes \n");
//    printf("        -s strat_sec  -- offset time from data head (sec)\n");
//    printf("                          (default :  0 )\n");
    printf("        -p period     -- data period to convert (sec)\n");
//    printf("                          (default : all)\n");
//    printf("        -r            -- reverse track order when fanout =2 or more\n");
//    printf("        -v vex_file   -- VEX file name to be read when \"information\n");
//    printf("                              file creation mode\" is set\n");
//    printf("        -sid stat_id  -- station ID when \"information file creation\n");
//    printf("                              mode\" is set\n");
//    printf("        -scan scan#   -- scan# for MODE get when \"information file\n");
//    printf("                              creation mode\" is set (default : 1)\n");
//    printf("        -subp subpass -- set subpass at information file\n");
//    printf("                              creation mode (default is \"A\")\n");
    printf("        -type1        -- outfile naming Type 1 (replace file extention to \"m5b\")\n");
    printf("        -pid thread_id -- set pickup thread ID compulsory\n");
    printf("        -ch1 nn     -- set VDIF CH number corresponding to Mark5B CH #1\n"); 
    printf("                          nn is VDIF channel number (1-16)\n"); 
    printf("        -ch2 nn     -- set VDIF CH number corresponding to Mark5B CH #2\n"); 
    printf("           ...........................\n"); 
    printf("        -ch16 nn    -- set VDIF CH number corresponding to Mark5B CH #16\n"); 
    printf("        -chall n1:n2:....:nNN\n");
    printf("                 -- set VDIF channels corresponding to Mark5 all channels\n"); 
    printf("                     n1  -- VDIF CH number (1-16) corresponding to Mark5 CH #1\n"); 
    printf("                     n2  -- VDIF CH number (1-16) corresponding to Mark5 CH #2\n"); 
    printf("                      ...............\n"); 
    printf("                     nNN --  maximum is 16 channels\n"); 
    printf("\n");
    printf("   Naming rule for Mark5B file created (VDIF.vdif : original VDIF name)\n");
    printf("        Type 0  VDIF.vdif.m5b (default)\n");
    printf("        Type 1  VDIF.m5b\n");
//    printf("   Enviroment variables\n");
//    printf("        %s    ---   default path to Mark5 data ( %s )\n",
//        M5DIR,getenv(M5DIR));
//    printf("        %s    ---   default path to Mark5 vex file ( %s )\n",
//        M5VEX,getenv(M5VEX));
    printf("\n");
}

void monit_runp(runp_t *runp)
// run parameter monitor
{
       printf(" ---------- RUN CONDITION ---------\n");
       printf("   vdifname  : %s\n",runp->vdifname);
       printf("   k5name    : %s\n",runp->m5name_given);
       printf("   outdir    : %s\n",runp->outdir);
       printf("   bitstream : %d\n",runp->nbitstream);
       printf("   fmhz      : %f\n",runp->fmhz);
       printf("   adbit     : %d\n",runp->adbit);
       printf("   k5numch   : %d\n",runp->k5numch);
       printf("   vdif_channels : %d %d %d %d\n",runp->chtable[0],
		   runp->chtable[1],runp->chtable[2],runp->chtable[3]);
       printf("   2bit mode : %d\n",runp->e2bit);
//       printf("   infofile  : %s\n",runp->info_file);
//       printf("   channel   : %d\n",runp->channel);
//       printf("   group     : %d\n",runp->group);
//       printf("   soffset   : %d\n",runp->soffset);
//       printf("   period    : %d\n",runp->period);
//       printf("   odd       : %d\n",runp->odd);
//       if(runp->vex_file[0] != '\0')
//          printf("   vex_file  : %s\n",runp->vex_file);
       if(runp->sid[0] != '\0' && runp->sid[0] !='-')
          printf("   station_id: %s\n",runp->sid);
       printf(" -------------------------------------\n");
    printf("\n");
}

int main(int argc, char *argv[])
{
       int i;
       double telapsed;           // to measure processing time
       char fname[256];
       char program_version[60];
	   time_t tstart, tfinish;    
       runp_t runp;

       if(argc == 1 || (argc == 2 && strncmp(*(argv+1),"--help",6)==0 )){ // 2008-02-19
          print_usage();
          exit(0);
       }
       if(argc == 2 && strncmp(*(argv+1),"--ver",5)==0 ){ // 2008-02-19
          printf("\n");
          printf("%s\n",PROGRAM_VERSION);
          printf("\n");
          exit(0);
       } 
       if (argc>=2){
          if(strncmp(*(argv+1),"env",3)==0){
          printf("   Enviroment variables\n");
          printf("        %s ---   default directory for Mark5 data\n",M5DIR);
          printf("                      ( %s )\n",getenv(M5DIR));
          printf("\n");
          exit(0);
          }
       }

       strcpy(program_version,PROGRAM_VERSION);

       /* run parameters get */
       param_get_vdif2m5b(argc, argv, &runp); 

       /* out directory create if not set */

       if(runp.outdir[0]=='\0'){
          parse_dir_and_name(runp.vdifname, runp.outdir, fname); // debug
          //printf("A %s\n",runp.mk5name);
          //printf("A %s\n",runp.outdir);
          //printf("A %s\n",fname);
       }


       /* Program Name Display */
       printf("************************************************************\n");
       printf("*    VDIF to Mark5B Data Format Converter                  *\n");
       printf("*      %32s  by  T.KONDO       *\n",program_version);
       printf("************************************************************\n");
       printf("\n");

//       monit_runp(&runp);

       tstart= time(NULL);


       i=convVDIF2M5B(&runp);

       if(i < 0) goto exitend;

       //printf("  K5 file (%s) has been created.\n\n",runp.k5name);


       tfinish= time(NULL);
       telapsed = difftime(tfinish,tstart);
       printf("\n Time elapsed (sec) for One obs process is %f\n",telapsed);

exitend: 
       return 0;
}


/* UTC to MJD conversion */
int ut2mjd(double *d, int year, int month, int day,
           int hour, int min, int sec)
/*
  [in]
     year   -- year (yyyy)
     month  -- month (month)
     day    -- day of month (dd)
     hour   -- hour (hh)
     min    -- minute (mm)
     sec    -- second (ss)
  [return]
     ut2mjs -- MJD  (integer day)
     d      -- fractional part of MJD (day)
*/
{
    int  bc, gregory, mnth, y4; 
    //long j;
    int j;

	mnth=month;
	y4=year;
    bc = year <= 0;
    
    /* First day at Gregory's calendar */
    gregory = year > 1582 || (year == 1582 && month > 10) ||
              (year == 1582 && month == 10 && day >= 15);
            
    if (mnth <= 2) {
        y4--;
        mnth += 12;
    }
    
    if (hour < 12) {
        j = 0;
        *d = 0.5;
    } else {
        j = 1;
        *d = -0.5;
    }
    *d += (hour*3600L + min*60 + sec)/86400.0;
    
    j += bc ? (y4-3)/4 : y4/4;
    if (gregory) j+= 2 - y4/100 + y4/400;
    j += 1720994L + y4*365L + (mnth+1)*30 + (mnth+1)*3/5 + day;
    
	/* convert to MJD */
	if (hour < 12) {
		*d -= 0.5;
	} else {
		*d += 0.5;
		j -= 1L;
	}
    j -= 2400000L;
    return j;
}

void ch_realloc(M5B_HEADER_T *m5b, int *chtable)
{
// supposed maximum bitstream 2bit x 16ch
//  Mark5B is 1bit or 2bit AD

    int i,n,m,numch,adbit,bs;
    int pch;
    unsigned int iw, iw32, ow32;
    
    unsigned int imask[4]={0x1, 0x3, 0xF, 0xFF};
    unsigned int umask[4]={0x7FFFFFFF, 0x3FFFFFFF, 0x0FFFFFFF, 0x00FFFFFF};

    adbit=m5b->si.adbit;
    numch=m5b->si.numch;
    bs=adbit*numch;

    // convert each 32-bit word
    for(i=0; i<m5b->ndata; i++){
        iw32=m5b->idata[i];
        ow32=0;
        for(n=0; n< 32/bs; n++){
            for(m=0; m<numch; m++){
                ow32 >>= adbit;
                pch=chtable[m];
                if(pch > numch) pch=m+1;  // doesn't change
                iw= iw32 >> adbit*(pch-1);
                iw &= imask[adbit-1];
                iw <<= 32-adbit;
                ow32= (ow32 & umask[adbit-1]) | iw;
            }
            iw32 >>= bs;
        }
        m5b->idata[i]=ow32;
    }
}



void disp_condition(runp_t *runp, vdif_file_t *vd)
{
/*
       display VDIF and conversion infomation
*/

    int n;
	printf("\n");
	//printf("========================= SUMMARY =============================\n");
	printf("===================== VDIF FILE INFO ===========================\n");
	printf(" File Name   : %s\n",runp->vdifname);
    printf("      Size   : %lld bytes\n",(long long)vd->f_size); 
	printf(" Start Time  : %04d/%02d/%02d %02d:%02d:%02d\n", 
        vd->vhead.t.year,vd->vhead.t.month,vd->vhead.t.mday,vd->vhead.t.hh,vd->vhead.t.mm,vd->vhead.t.ss);
    printf("               Thread ID : %d      Frame# : %d\n", vd->vhead.thread_id,vd->vhead.frame);
    printf("               Pickup Thread ID : ");
    if(runp->pickup_tid >0){
        printf("%d\n",runp->pickup_tid);
    } else {
        printf("%d\n",vd->vhead.thread_id);
    }
	printf(" Bytes/Frame : %d bytes\n", vd->vhead.frame_len64*8);
	if(vd->vhead.d_type==0){
	    printf(" Data type   : Real Data    ");
    } else {
	    printf(" Data type   : Complex Data ");
    }
    printf("    AD bits : %d", vd->vhead.bpsampl+1);
    printf("    #Channels : %d\n", (int)pow(2.0, vd->vhead.log2_chs));
	printf(" EDV#        : %d\n",vd->vhead.edv);
	if(vd->vhead.edv==1){   // This is NICT's EDV
       printf(" Extended header information (NICT\'s EDV)\n");
	   printf("      Sampling Frequency : %d", vd->vhead.srate);
       if(vd->vhead.uflag==0){
		   printf(" kHz\n");
	   } else {
		   printf(" MHz\n");
	   }
	   printf("      Sync block         : %08X\n", vd->vhead.sync_block);
	   printf("      DAS/Station name   : %s\n", vd->vhead.das);
	}
//	printf(" # of frames/sec     : %d\n",ip->framsec);     
//	printf(" # of total frames   : %d",ip->totalframes);     
//	printf("   (#bad frames : %d)\n",ip->badframes);     
//	printf(" Data period (sec)   : %f\n",ip->dlength);     
//	printf(" Data Rate/thread (Mbps) : %f\n",ip->fbps/1.0e6);     
	printf(" Sampling Frequency Estimated (MHz) : %f\n",vd->fsample/1.0e6); 
    if(runp->flag_ch_alloc==1){
        printf("\n");
        n=0;
        printf(" Channel re-allocation: ");
        while(n < vd->numch){
        //while(n < 16){
            if(n > 0 && (n % 4)==0) printf("\n                        ");            
            printf(" (%2d - %2d)",n+1,runp->chtable[n]);
            n++; 
        }
        printf("\n");
    }


	printf("=================================================================\n");

	printf("\n");

}



int fwdNextSecondDIF(vdif_file_t *vd)
{
/* 
       forward next on the second with in the same thread ID
	   if current position is not on the second,
	   1sec forward means forward at the begining of the next second
	   (so less than 1 sec forward)
*/
	int64_t pos,posk;
    int r_thread_id;

    pos=vd->pos;  // save current file pointer
    r_thread_id=vd->vhead.thread_id;

	// compute set position

    posk=pos+vd->vhead.frame_len64;
    
	while(posk < vd->f_size){  // 
		fseek(vd->fin,posk,SEEK_SET); // move pointer by K5->bytesec
        if( ferror( vd->fin ) ){
            return -1;
        }
		vd->pos=posk;
	    /* get headder info */
        if(vdif_header_read_mode1(vd)<0) return -1;
        if(vd->vhead.frame == 0 && vd->vhead.thread_id==r_thread_id) goto exit2return;
    }
exit2return:       
    return 0;
}



int convVDIF2M5B(runp_t *runp)
/*
    VDIF to Mark5B conversion  2016-12-20 
	     (support only 1 ch mode)
*/
{
    int pcount,icount,opos,ipos;
    int i,n,k,m,m4;
    int iwrite;
    int ierr;
	int frame_c;   // total frame numbervcounter 2012-10-17
    int nw;
    unsigned int iw, iw32;
    int64_t totalbyte;
	//int seconds;
    //FILE *fo;
    FILE *fo;
    vdif_file_t vd;
    timeval_vdif_t time1st;
    int r_thread_id;
//    vdifconv_t ct;
	int *conv2bit;
    int adbit;
	int conv2bit_1[]={0,1,2,3};   // offset binary (no conversion 2 bit)
    int conv2bit_2[]={0,2,1,3};   //  conversion 2 bit (default)
    int conv2bit_3[]={1,0,2,3};   //  conversion 2 bit singed integer

    //VDIF_HEADER_T vdif;
    M5B_HEADER_T m5b;
    int seconds_init; // start time in seconds
    int mjd1st;
    double dw,dsec,dmsec10;

    int sec_count;



    //printf("  convVDIF2M5B : \n");
    ierr=0;
	frame_c=0; 

    //head=obuf;
    //idata=obuf+4;
    m5b.head=m5b.ibuf;
    m5b.idata=m5b.ibuf+4;
    m5b.nbuf=2504;
    m5b.ndata=2500;

	if(runp->e2bit==1){
	   //ct.conv2bit=conv2bit_1;
	   conv2bit=conv2bit_1;
	} else if(runp->e2bit==3){
	   //ct.conv2bit=conv2bit_3;
	   conv2bit=conv2bit_3;
	} else {
	   //ct.conv2bit=conv2bit_2;
	   conv2bit=conv2bit_2;
	}

	strcpy(vd.fname,runp->vdifname);
	vd.skip_bytes=runp->skip_bytes;
	vd.flag_compulsory=runp->flag_compulsory;

    /* get VDIF file info by opening and get reading file (then close) */
    if(vdif_t_ini(&vd,0)<0) goto exit2return;   // VDIF data check with auto mode
    //if(vdif_t_ini(&vd,1)<0) goto exit2return;  // VDIF data check with 1 sec reading compulsory

    if(vd.fsample==0.0){
        if(runp->fmhz==0.0){
            printf("Warning!  sampling frequency is not included in VDIF headder!!\n");
            printf("Please set sampling frequency using \"-fsampl fMHz\" option!!\n");
            exit(0);
        } else {
            vd.fsample=1.0e6*runp->fmhz;
        }
    } else {
        if(runp->fmhz != 0.0){  // compulsory set sampling frequency
            vd.fsample=1.0e6*runp->fmhz;
        }
    }

    // check channel re-allocation condition 
    if(runp->flag_ch_alloc==1){
        if((32 % (vd.adbit*vd.numch)) != 0){
            printf(" Warning!! Channel re-allocation condition is not satified!!\n");
            printf("           so channel is not re-allocated!!\n");
            runp->flag_ch_alloc=0;
        }
    }

    disp_condition(runp, &vd);

    if(vd.vhead.frame != 0){
        printf(" Warning!! VDIF data don't start at on the second!!\n");
        printf("           so forward to the next on the second!!\n");
        if(fwdNextSecondDIF(&vd) < 0) {
            printf("  Error at forwarding!\n");
            exit(0);
        }
	    time1st=vd.vhead.t;
    }
	
    time1st=vd.vhead.t;  // save 1st time 

//    printf("  VDIF start time :  %4d/%02d/%02d %02d:%02d:%02d  FRAME# %d THREAD_ID# %d\n",
//        vd.vhead.t.year,vd.vhead.t.month,vd.vhead.t.mday,vd.vhead.t.hh,vd.vhead.t.mm,vd.vhead.t.ss,
//        vd.vhead.frame,vd.vhead.thread_id);

	/* 
	  set reading buffer size to satisfy following condition
	     1. (frame size)/(buffer size) should be an integer number
	*/
	vd.nword=vd.frame_bytes/4;     
	while (vd.nword > BUFMAX) vd.nword/=2;

	if(((vd.frame_bytes/4) % vd.nword) != 0){
		printf(" convVDIF2MK5B: buffer size does\'nt satisfy the condition! %d %d\n",vd.nword, vd.frame_bytes*4);
	}

	vd.imax=(vd.frame_bytes/4)/vd.nword;   // # of loops to read entire one frame data
	vd.nsample=vd.nword*32/vd.adbit;  // # of samples in buffer (don't care about # of Chs)

    vd.ubuf  = (u_int32_t *)malloc(vd.nword*sizeof(u_int32_t)); //  data read buffer

	vd.imax_c=0;

	/* VDIF file open for conversion */
	if( (vd.fin  = fopen( vd.fname, "rb" )) == NULL ){
         printf("File (%s) Open Error! \n",vd.fname);
         goto exit2return;
    }
	vd.flag_open=1;
    //printf("  VDIF Data File  : %s\n",vd.fname);

	if(vd.flag_vdif_anal==1){ // read pointer move to 1st valid header
		fseek(vd.fin,vd.pos_1st_header,SEEK_SET); // move pointer by K5->bytesec
	}

    /* foward startoffset seconds */
	if(runp->soffset>0){
		printf("  %d second forwading.....",runp->soffset);
	    if(fwdNsecVDIF(runp->soffset, &vd) <0){
           printf("Foward error %d\n",runp->soffset);
           i=fclose( vd.fin ) ;
           goto exit2return;
        }
	    printf("\n");
	    /* get headder info again */
        if(vdif_header_read_mode1(&vd)<0) goto exit2return;
	    time1st=vd.vhead.t;
	}

	r_thread_id=vd.vhead.thread_id;  // set reading thread ID to 1st one 
    if(runp->pickup_tid > 0) r_thread_id=runp->pickup_tid;


    /* output file create */
    if(runp->m5name[0]=='\0'){
       m5bname_make(runp);
    }
    printf("\n");
    printf(" Mark5B Data File will be created : %s\n",runp->m5name);
    if( (fo  = fopen( runp->m5name, "wb" )) == NULL ){
           printf("      File  %s  Open Error! \n",runp->m5name);
           ierr= -2;
           goto exit2return;
    }

    pcount=0; // converted data period counter (in sec)

    /* header monitor */
    printf(" ==============================================================================\n");
	printf("                                                             EXTENDED DATA\n");
	printf(" HH:MM:SS  FRAME#  I  L  VER  CHS     #Bytes  C AD  TID SID       W1       W2\n");
    printf(" ------------------------------------------------------------------------------\n");

    /* Mark5 header common info setup */

    m5b.t.mjd=ut2mjd(&dw,vd.vhead.t.year,vd.vhead.t.month,vd.vhead.t.mday,0,0,0);
    m5b.t.mjd=m5b.t.mjd % 1000;
    mjd1st=m5b.t.mjd;
    m5b.t.msec10=0;
    seconds_init=vd.vhead.t.second;
    m5b.t.second=seconds_init;
    m5b.tvg=0;
    m5b.uid=0xF00F;
    dsec=0.0;
    dmsec10=0.0;
    sec_count=0;

    //nframe=0;  // frame # (in a second) counter reset
    m5b.frame=0;  // frame # (in a second) counter reset


    m5b.fsample=vd.fsample;
    m5b.p_frame=(double)((m5b.ndata*32)/(vd.adbit*vd.numch))/m5b.fsample;  // period/frame
    if(((m5b.ndata*32) % (vd.adbit*vd.numch)) !=0 ){
        printf("conVDIF2M5B:  Error!!  (cannot convert to Mark5B format)\n");
        exit(0);
    }
    adbit=vd.adbit;

    m5b.si.adbit=vd.adbit;
    m5b.si.numch=vd.numch;
    m5b.si.sfkHz=(int)(m5b.fsample/1.0e3 + 0.5);

    /* Mark5B initial header create */
       /* Mark5B header generation */
    make_m5b_header(&m5b);

    //ct.nbit=0;
    icount=0;  // 32bit word counter
    opos=0;    // out buffer writing position

    totalbyte=0;

    while(1){

        m5b.frame=0;
		/* second loop */
		for(n=0; n<vd.nframsec; n++){
		   /* VDIF header read */
		   i=vdif_header_read_mode2(&vd);

           if(i<0) goto exithere;
           // check thread ID
           if(vd.vhead.thread_id != r_thread_id){  // next headder read
               // move read pointer by 1 frame
               fseek(vd.fin,vd.frame_bytes,SEEK_CUR);
               continue;
           }

		   if(n==0){
		        printf(" %02d:%02d:%02d %6d",vd.vhead.t.hh,vd.vhead.t.mm,vd.vhead.t.ss, vd.vhead.frame);
	            printf(" %3d %2d  %3d   %2d %10d  %1d %2d %4d  %s %08X %08X\n", 
				vd.vhead.invalid,vd.vhead.legacy,vd.vhead.ver,
					     (int)pow(2.0, vd.vhead.log2_chs),vd.vhead.frame_len64*8,
					     vd.vhead.d_type,vd.vhead.bpsampl+1,vd.vhead.thread_id,vd.vhead.sid,vd.vhead.w1,vd.vhead.w2);
		   }
		   /* VDIF data 1 frame read */
		   for(k=0; k<vd.imax; k++){
			   vd.wordread = fread( vd.ubuf, sizeof( unsigned int ), vd.nword, vd.fin );
			   if (vd.wordread ==0 ) { // EOF
                   goto exithere; 
		       }
               // copy data to m5b buffer and write to file
               ipos=0;
               //nw=m5b.ndata-opos;
               if(m5b.ndata <= vd.wordread){   // CASE1: m5b.ndata <= vd.wordread
                   while(ipos+m5b.ndata <= vd.wordread){
                       // copy VDIF data to M5B buffer
                       nw=m5b.ndata-opos;
                       memcpy(m5b.idata+opos,vd.ubuf+ipos,nw*sizeof(int));
                       opos+=nw;
                       ipos+=m5b.ndata;
                       if(opos >= m5b.ndata){
                          // M5B frame filled

                          // encode conversion
                          if(adbit==2){
                              for(m=0; m<m5b.ndata; m++){
                                  iw32=m5b.idata[m];
                                  for(m4=0; m4<16; m4++){
                                     iw=(iw32 & 0x3);
                                     iw=conv2bit[iw];
                                     iw32 >>=2;
                                     iw32=((iw32 & 0x3FFFFFFF) | (iw <<= (32-adbit)));
                                  }
                                  m5b.idata[m]=iw32;
                              }
                          }
                          if(runp->flag_ch_alloc==1){
                             ch_realloc(&m5b,runp->chtable);
                          }

                          iwrite=fwrite(m5b.ibuf,sizeof(int),m5b.nbuf,fo);  // write one frame data including header
                          totalbyte+=m5b.nbuf*sizeof(int);
                          m5b.frame++;
                          // time update for Mark5B header
                          m5b_time_update(&m5b, mjd1st, seconds_init, sec_count);
                          make_m5b_header(&m5b);    // header update
                          opos=0;
                       }
                   }
                   nw=vd.wordread-ipos;
                   if(nw > 0){
                       printf("convVDIF2M5B:  copy rest %d 32-bit words.\n",nw);
                       memcpy(m5b.idata+opos,vd.ubuf+ipos,nw*sizeof(int));
                       opos+=nw;
                   }
               } else {   // CASE2: m5b.ndata > vd.wordread

                   if(opos+vd.wordread < m5b.ndata){
                        memcpy(m5b.idata+opos,vd.ubuf+ipos,vd.wordread*sizeof(int));
                        opos+=vd.wordread;
                        ipos=0;
                   } else {
                       nw=m5b.ndata-opos;
                       memcpy(m5b.idata+opos,vd.ubuf+ipos,nw*sizeof(int));
                       ipos+=nw;

                       // M5B frame filled

                       // encode conversion
                       if(adbit==2){
                          for(m=0; m<m5b.ndata; m++){
                              iw32=m5b.idata[m];
                              for(m4=0; m4<16; m4++){
                                 iw=(iw32 & 0x3);
                                 iw=conv2bit[iw];
                                 iw32 >>=2;
                                 iw32=((iw32 & 0x3FFFFFFF) | (iw <<= (32-adbit)));
                              }
                              m5b.idata[m]=iw32;
                          }
                       }
                       // channel reallocation if set so
                       if(runp->flag_ch_alloc==1){
                           ch_realloc(&m5b,runp->chtable);
                       }

                       iwrite=fwrite(m5b.ibuf,sizeof(int),m5b.nbuf,fo);  // write one frame data including header
                       totalbyte+=m5b.nbuf*sizeof(int);
                       m5b.frame++;
                       // time update for Mark5B header
                       m5b_time_update(&m5b, mjd1st, seconds_init, sec_count);
                       make_m5b_header(&m5b);    // header update
                       opos=0;
                       if(ipos <= vd.wordread){
                           memcpy(m5b.idata+opos,vd.ubuf+ipos,(vd.wordread-ipos)*sizeof(int));
                           opos+=(vd.wordread-ipos);
                           ipos=0;
                       }
                   }


               }
           }
		   frame_c++;  // VDIF frame counter update
		}
        // rest write
        if(opos > 0){
             printf("convVDIF2M5B:  write rest %d 32-bit words.\n",opos);
             iwrite=fwrite(m5b.ibuf,sizeof(int),opos+4,fo);
             totalbyte+=(opos+4)*sizeof(int);
             opos=0;
        }
		//seconds++;
        sec_count++;
        // time update for Mark5B header
        m5b_time_update(&m5b, mjd1st, seconds_init, sec_count);
        m5b.frame=0;
        make_m5b_header(&m5b);    // header update

        // period check
        if(runp->period > 0){
            if(sec_count >= runp->period) goto exithere;
        }
	}
exithere:
    if(vd.flag_open==1) i=fclose( vd.fin) ;
    vd.flag_open=0;
    // rest of data write
    if(opos != 0){
       printf("convVDIF2M5B:  write final %d 32-bit words.\n",opos);
       iwrite=fwrite(m5b.ibuf,sizeof(int),opos+4,fo);
       totalbyte+=(opos+4)*sizeof(int);
    }
    if(totalbyte==0) printf("  NO DATA WERE CONVERTED!!\n");
    printf(" ==============================================================================\n");
	printf("\n");
    i=fclose( fo ) ;

exit2return:
       //printf("  K5 file (%s) has been created.\n\n",runp->k5name);
	free(vd.ubuf);
    if(vd.flag_open==1) i=fclose(vd.fin);

    printf("\n");
//    if(ct.k5nfile==1){
    printf(" Mark5B Data File created : %s\n",runp->m5name);
    if(totalbyte ==0){
        printf("                             but contents are empty!!\n");
    } else {
        printf("                            (%lld bytes)\n",(long long)totalbyte);
    }

    return 0;
}


void m5bname_make(runp_t *rp)
{
/*
  Create M5b file name  
         2016-12-20
*/
    int i,len;
    //int flag_remove=0;
    int flag_remove;
    char m5dir[256],dir[256], name[256],fname[256]; 
    char dir2[256],name2[256];

    flag_remove=0;
    if(rp->naming_type==1) flag_remove=1;

    //parse_dir_and_name(adsname, dir, name);
    parse_dir_and_name(rp->vdifname, dir, name);

    /* remove file extension if set so */

    if(flag_remove==1){
        len=strlen(name);

         // find last "." in the name to remove extension
        for(i=len-1; i>=0; i--){
           if(name[i]=='.'){
               name[i]='\0';
               break;
           }
        }
    }


    // out directory prepare
    if(rp->outdir[0] == '\0'){
        m5dir[0]='\0';
    } else {
        // converting '\' to '/' if so
        len=strlen(rp->outdir);
        //printf("C %d %s\n",len,outdir);
        for(i=0; i<len; i++){
            if(rp->outdir[i]=='\\'){
                m5dir[i]='/';
            } else {
                m5dir[i]=rp->outdir[i];
            }
        }
        m5dir[len]='\0';
        // checkdir name accompanied by "/" or not 
        if(m5dir[len-1] != '/') {
           m5dir[len]='/';
           m5dir[len+1]='\0';
        }
    }

    // check given m5name is accompanied by directory or not
    dir2[0]='\0'; name2[0]='\0';
    if(rp->m5name_given[0] != '\0'){
        parse_dir_and_name(rp->m5name_given, dir2, name2);
    }

    // create m5bname

    if(rp->m5name_given[0] == '\0'){  // in case ofnot given from an operator
       strcpy(fname,name);
       strcat(fname,".m5b");
       strcpy(rp->m5name,m5dir);
       strcat(rp->m5name,fname);
    } else {   // in case of given from an operator
        if(dir2[0]=='\0'){
            if(rp->outdir[0]=='\0'){ 
                strcpy(rp->m5name,rp->m5name_given);
            } else {
                strcpy(rp->m5name,m5dir);
                strcat(rp->m5name,name2);
            }
        } else {
            strcpy(rp->m5name,rp->m5name_given);
        }
    }
}


//void m5b_hd_make(int *head, int mjd, int seconds, int msec10,
//				 int uid, int tvgflag, int nframe)
void make_m5b_header(M5B_HEADER_T *m5b)
/*
     Mark5B header block creation
     2009-06-07 

  [in]
      mjd      --  MJD (3digits)
	  seconds  --- seconds in day
	  msec10   --- msec*10
	  uid      --   User-specified field  (lower 16-bits are valid)
	  tvgflag  --   Tvg data flag   1: tvg data  0: not tvg data
	  nframe   --   Frame# within second (starting at 0)]
  [out]
      head  ---- pointer to header array (4 X 32 bits) buffer 
*/
{
    /* Mark 5B header block creation */
    int n,a;
    int crc16;
    //int64_t lw1;
    int idat[12];
    unsigned char bdat[6];
    int sync_pat=0xABADDEED;   // Mark5B sync pattern


    /* Sync word   */
	m5b->head[0]=sync_pat;

	/* User-specfied field (16 bits) */

	m5b->head[1]=(m5b->uid & 0xFFFF) << 16;

	/* TVG flag */
	m5b->head[1]=((m5b->head[1] & 0xFFFF0000) | ((m5b->tvg & 0x1) << 15));

	/* Frame # field */
	//head[1]=((head[1] & 0xFFFF8000) | (nframe & 0x7FFF));
	m5b->head[1]=((m5b->head[1] & 0xFFFF8000) | (m5b->frame & 0x7FFF));

    /* Time field */

    //  JJJSSSSSssss

    a=100;
    for(n=0; n<3; n++){
        idat[n]=(m5b->t.mjd/a) % 10;
        a/=10;
    }
    a=10000;
    for(n=3; n<8; n++){
        idat[n]=(m5b->t.second/a) % 10;
        a/=10;
    }
    a=1000;
    for(n=8; n<12; n++){
        idat[n]=(m5b->t.msec10/a) % 10;
        a/=10;
    }

	m5b->head[2]=idat[0] & 0xF;
    for(n=1; n<8; n++){  // 4 bit step
		m5b->head[2]=((m5b->head[2] << 4) | (idat[n] & 0xF));
	}

	m5b->head[3]=idat[8] & 0xF;
    for(n=9; n<12; n++){  // 4 bit step
		m5b->head[3]=((m5b->head[3] << 4) | (idat[n] & 0xF));
	}
	m5b->head[3] <<= 16;

    /* CRC16 */   
    for(n=0; n<6; n++){
        bdat[n]= (idat[2*n] << 4) + idat[2*n+1];
    }
	crc16 = crcc(bdat, 48, 040003, 16);
    //printf("Debug CRC16 %08X\n",rvsint(crc16));
    //printf("Debug CRC16 %08X\n",crc16);

	// Mark5B bit order is reversal 
	crc16=rvsint(crc16) >> 16;


	m5b->head[3] |= (crc16 & 0xFFFF);
    
}



/* crc subroutine ripoff  by David*/
int crcc(unsigned char* idata, int len, int mask, int cycl) {
	unsigned int istate = 0;
	int idbit;
	int q, ich, icb;

	for (idbit = 1; idbit <= len; idbit++) {
		q = istate & 1;
		ich = (idbit - 1) / 8;
		icb = 7 - (idbit - 1) % 8;
		if ((((idata[ich] >> icb) & 1) ^ q) == 0)
			istate &= -2;
		else {
			istate ^= mask;
			istate |= 1;
		}
		istate = (istate >> 1) | ((istate & 1) << (cycl -1));  
	}

	return ((int) istate);
}


void m5b_time_update(M5B_HEADER_T *m5b, int mjd1st, int seconds_init, int sec_count)
{
   // time update for Mark5B 
    double dsec,dmsec10;
    dsec=(double)m5b->frame*m5b->p_frame;   // fractional second
    dmsec10=dsec*10000.0;
    m5b->t.msec10=(int)dmsec10;
    m5b->t.second=seconds_init+sec_count;
    m5b->t.mjd=mjd1st+(int)((double)m5b->t.second/86400.0);
    if(m5b->t.second >= 86400) m5b->t.second-=86400;
}